@@ -2,10 +2,13 @@
 
 const Big = imports.gi.Big;
 const Clutter = imports.gi.Clutter;
+const GLib = imports.gi.GLib;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const Main = imports.ui.main;
 
@@ -33,9 +36,16 @@ RunDialog.prototype = {
 
         this._internalCommands = { 'lg':
                                    Lang.bind(this, function() {
+                                       // Run in an idle to avoid recursive key grab problems
                                        Mainloop.idle_add(function() { Main.createLookingGlass().open(); });
                                    }),
-                                   
+
+                                   'r': Lang.bind(this, function() {
+                                       let global = Shell.Global.get();
+                                       global.reexec_self();
+                                   }),
+
+                                   // Developer brain backwards compatibility
                                    'restart': Lang.bind(this, function() {
                                        let global = Shell.Global.get();
                                        global.reexec_self();
@@ -69,7 +79,7 @@ RunDialog.prototype = {
 
         let label = new Clutter.Text({ color: BOX_TEXT_COLOR,
                                        font_name: '18px Sans',
-                                       text: 'Please enter a command:' });
+                                       text: _("Please enter a command:") });
         label.set_position(6, 6);
         boxGroup.add_actor(label);
 
@@ -106,12 +116,13 @@ RunDialog.prototype = {
         if (f) {
             f();
         } else if (command) {
-            var p = new Shell.Process({'args' : [command]});
             try {
+                let [ok, len, args] = GLib.shell_parse_argv(command);
+                let p = new Shell.Process({'args' : args});
                 p.run();
             } catch (e) {
                 // TODO: Give the user direct feedback.
-                log('Could not run command ' + command + '.');
+                log('Could not run command ' + command + ': ' + e);
             }
         }
     },
@@ -120,7 +131,7 @@ RunDialog.prototype = {
         if (this._isOpen) // Already shown
             return;
 
-        if (!Main.startModal())
+        if (!Main.beginModal())
             return;
             
         this._isOpen = true;
