@@ -11,11 +11,9 @@ const Pango = imports.gi.Pango;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 
-const AppDisplay = imports.ui.appDisplay;
 const DND = imports.ui.dnd;
-const GenericDisplay = imports.ui.genericDisplay;
 const Main = imports.ui.main;
-const Overlay = imports.ui.overlay;
+const Overview = imports.ui.overview;
 const Panel = imports.ui.panel;
 const Tweener = imports.ui.tweener;
 
@@ -274,6 +272,12 @@ Workspace.prototype = {
 
         this.actor = new Clutter.Group();
         this.actor._delegate = this;
+        // Auto-sizing is unreliable in the presence of ClutterClone, so rather than
+        // implicitly counting on the workspace actor to be sized to the size of the
+        // included desktop actor clone, set the size explicitly to the screen size.
+        // See http://bugzilla.openedhand.com/show_bug.cgi?id=1755
+        this.actor.width = global.screen_width;
+        this.actor.height = global.screen_height;
         this.scale = 1.0;
 
         let windows = global.get_windows().filter(this._isMyWindow, this);
@@ -293,16 +297,16 @@ Workspace.prototype = {
                               Lang.bind(this,
                                         function(clone, time) {
                                             this._metaWorkspace.activate(time);
-                                            Main.overlay.hide();
+                                            Main.overview.hide();
                                         }));
         this.actor.add_actor(this._desktop.actor);
 
         // Create clones for remaining windows that should be
-        // visible in the overlay
+        // visible in the Overview
         this._windows = [this._desktop];
         this._windowIcons = [ null ];
         for (let i = 0; i < windows.length; i++) {
-            if (this._isOverlayWindow(windows[i])) {
+            if (this._isOverviewWindow(windows[i])) {
                 this._addWindowClone(windows[i]);
             }
         }
@@ -318,7 +322,7 @@ Workspace.prototype = {
 
         this._frame = null;
 
-        this.leavingOverlay = false;
+        this.leavingOverview = false;
     },
 
     updateRemovable : function() {
@@ -346,7 +350,7 @@ Workspace.prototype = {
                 this._removeButton.set_opacity(0);
                 Tweener.addTween(this._removeButton,
                                  { opacity: 255,
-                                   time: Overlay.ANIMATION_TIME,
+                                   time: Overview.ANIMATION_TIME,
                                    transition: "easeOutQuad"
                                  });
             }
@@ -357,7 +361,7 @@ Workspace.prototype = {
             if (this._visible) {
                 Tweener.addTween(this._removeButton,
                                  { opacity: 0,
-                                   time: Overlay.ANIMATION_TIME,
+                                   time: Overview.ANIMATION_TIME,
                                    transition: "easeOutQuad",
                                    onComplete: this._removeRemoveButton,
                                    onCompleteScope: this
@@ -398,7 +402,9 @@ Workspace.prototype = {
 
     // Mark the workspace selected/not-selected
     setSelected : function(selected) {
-        if (selected) {
+        let global = Shell.Global.get();
+        // Don't draw a frame if we only have one workspace
+        if (selected && global.screen.n_workspaces > 1) {
             if (this._frame)
                 return;
 
@@ -428,7 +434,7 @@ Workspace.prototype = {
                              this._desktop.actor.height + 2 * FRAME_SIZE / this.actor.scale_y);
     },
 
-    // Reposition all windows in their zoomed-to-overlay position. if workspaceZooming
+    // Reposition all windows in their zoomed-to-Overview position. if workspaceZooming
     // is true, then the workspace is moving at the same time and we need to take
     // that into account
     positionWindows : function(workspaceZooming) {
@@ -460,7 +466,7 @@ Workspace.prototype = {
                                scale_x: scale,
                                scale_y: scale,
                                workspace_relative: workspaceZooming ? this : null,
-                               time: Overlay.ANIMATION_TIME,
+                               time: Overview.ANIMATION_TIME,
                                transition: "easeOutQuad",
                                onComplete: Lang.bind(this, function() {
                                   this._fadeInWindowIcon(clone, icon);
@@ -494,7 +500,7 @@ Workspace.prototype = {
         icon.raise(this.actor);
         Tweener.addTween(icon,
                          { opacity: 255,
-                           time: Overlay.ANIMATION_TIME,
+                           time: Overview.ANIMATION_TIME,
                            transition: "easeOutQuad" });
     },
 
@@ -531,14 +537,14 @@ Workspace.prototype = {
 
         // If metaWin.get_compositor_private() returned non-NULL, that
         // means the window still exists (and is just being moved to
-        // another workspace or something), so set its overlayHint
+        // another workspace or something), so set its overviewHint
         // accordingly. (If it returned NULL, then the window is being
         // destroyed; we'd like to animate this, but it's too late at
         // this point.)
         if (win) {
             let [stageX, stageY] = clone.actor.get_transformed_position();
             let [stageWidth, stageHeight] = clone.actor.get_transformed_size();
-            win._overlayHint = {
+            win._overviewHint = {
                 x: stageX,
                 y: stageY,
                 scale: stageWidth / clone.actor.width
@@ -552,7 +558,7 @@ Workspace.prototype = {
     },
 
     _windowAdded : function(metaWorkspace, metaWin) {
-        if (this.leavingOverlay) 
+        if (this.leavingOverview)
             return;
 
         let win = metaWin.get_compositor_private();
@@ -569,16 +575,16 @@ Workspace.prototype = {
             return;
         }
         
-        if (!this._isOverlayWindow(win))
+        if (!this._isOverviewWindow(win))
             return;        
 
         let clone = this._addWindowClone(win);
 
-        if (win._overlayHint) {
-            let x = (win._overlayHint.x - this.actor.x) / this.scale;
-            let y = (win._overlayHint.y - this.actor.y) / this.scale;
-            let scale = win._overlayHint.scale / this.scale;
-            delete win._overlayHint;
+        if (win._overviewHint) {
+            let x = (win._overviewHint.x - this.actor.x) / this.scale;
+            let y = (win._overviewHint.y - this.actor.y) / this.scale;
+            let scale = win._overviewHint.scale / this.scale;
+            delete win._overviewHint;
 
             clone.actor.set_position (x, y);
             clone.actor.set_scale (scale, scale);
@@ -588,48 +594,50 @@ Workspace.prototype = {
         this.updateRemovable();
     },
 
-    // Animate the full-screen to overlay transition.
-    zoomToOverlay : function() {
-        // Move the workspace into size/position
-        this.actor.set_position(this.fullSizeX, this.fullSizeY);
-        
-        this.updateInOverlay();
+    // Animate the full-screen to Overview transition.
+    zoomToOverview : function() {
+        this.actor.set_position(this.gridX, this.gridY);
+        this.actor.set_scale(this.scale, this.scale);
+
+        // Position and scale the windows.
+        this.positionWindows(true);
+
+        // Fade in the remove button if available, so that it doesn't appear
+        // too abrubtly and doesn't start at a too big size.
+        if (this._removeButton) {
+            Tweener.removeTweens(this._removeButton);
+            this._removeButton.opacity = 0;
+            Tweener.addTween(this._removeButton,
+                             { opacity: 255,
+                               time: Overview.ANIMATION_TIME,
+                               transition: 'easeOutQuad'
+                             });
+        }
 
         this._visible = true;
     },
 
-    // Animates the display of a workspace and its windows to have the current dimensions and position.
-    updateInOverlay : function() {
-        Tweener.addTween(this.actor,
-                         { x: this.gridX,
-                           y: this.gridY,
-                           scale_x: this.scale,
-                           scale_y: this.scale,
-                           time: Overlay.ANIMATION_TIME,
-                           transition: "easeOutQuad"
-                         });
-
-        // Likewise for each of the windows in the workspace.
-        this.positionWindows(true);
-    },
-
-    // Animates the return from overlay mode
-    zoomFromOverlay : function() {
-        this.leavingOverlay = true;
+    // Animates the return from Overview mode
+    zoomFromOverview : function() {
+        this.leavingOverview = true;
 
         this._hideAllIcons();
 
-        Tweener.addTween(this.actor,
-                         { x: this.fullSizeX,
-                           y: this.fullSizeY,
-                           scale_x: 1.0,
-                           scale_y: 1.0,
-                           time: Overlay.ANIMATION_TIME,
-                           transition: "easeOutQuad",
-                           onComplete: this._doneLeavingOverlay,
-                           onCompleteScope: this
-                         });
+        Main.overview.connect('hidden', Lang.bind(this,
+                                                 this._doneLeavingOverview));
 
+        // Fade out the remove button if available, so that it doesn't
+        // disappear too abrubtly and doesn't become too big.
+        if (this._removeButton) {
+            Tweener.removeTweens(this._removeButton);
+            Tweener.addTween(this._removeButton,
+                             { opacity: 0,
+                               time: Overview.ANIMATION_TIME,
+                               transition: 'easeOutQuad'
+                             });
+        }
+
+        // Position and scale the windows.
         for (let i = 1; i < this._windows.length; i++) {
             let clone = this._windows[i];
             Tweener.addTween(clone.actor,
@@ -638,7 +646,7 @@ Workspace.prototype = {
                                scale_x: 1.0,
                                scale_y: 1.0,
                                workspace_relative: this,
-                               time: Overlay.ANIMATION_TIME,
+                               time: Overview.ANIMATION_TIME,
                                opacity: 255,
                                transition: "easeOutQuad"
                              });
@@ -656,7 +664,7 @@ Workspace.prototype = {
                            y: this.gridY,
                            scale_x: this.scale,
                            scale_y: this.scale,
-                           time: Overlay.ANIMATION_TIME,
+                           time: Overview.ANIMATION_TIME,
                            transition: "easeOutQuad",
                            onComplete: Lang.bind(this, this._fadeInAllIcons)
                          });
@@ -678,7 +686,7 @@ Workspace.prototype = {
                            y: this.gridY,
                            scale_x: this.scale,
                            scale_y: this.scale,
-                           time: Overlay.ANIMATION_TIME,
+                           time: Overview.ANIMATION_TIME,
                            transition: "easeOutQuad"
                          });
 
@@ -701,7 +709,7 @@ Workspace.prototype = {
                            y: destY,
                            scale_x: this.scale,
                            scale_y: this.scale,
-                           time: Overlay.ANIMATION_TIME,
+                           time: Overview.ANIMATION_TIME,
                            transition: "easeOutQuad",
                            onComplete: onComplete
                          });
@@ -724,9 +732,9 @@ Workspace.prototype = {
         this._metaWorkspace.disconnect(this._windowRemovedId);
     },
 
-    // Sets this.leavingOverlay flag to false.
-    _doneLeavingOverlay : function() {
-        this.leavingOverlay = false;
+    // Sets this.leavingOverview flag to false.
+    _doneLeavingOverview : function() {
+        this.leavingOverview = false;
     },
 
     // Tests if @win belongs to this workspaces
@@ -735,13 +743,10 @@ Workspace.prototype = {
             (win.get_meta_window() && win.get_meta_window().is_on_all_workspaces());
     },
 
-    // Tests if @win should be shown in the overlay
-    _isOverlayWindow : function (win) {
-        let wintype = win.get_window_type();
-        if (wintype == Meta.WindowType.DESKTOP || 
-            wintype == Meta.WindowType.DOCK)
-            return false;
-        return !win.is_override_redirect();
+    // Tests if @win should be shown in the Overview
+    _isOverviewWindow : function (win) {
+        let appMon = Shell.AppMonitor.get_default()
+        return appMon.is_window_usage_tracked(win.get_meta_window());
     },
 
     _createWindowIcon: function(window) {
@@ -811,7 +816,7 @@ Workspace.prototype = {
     },
 
     _onCloneSelected : function (clone, time) {
-        Main.overlay.activateWindow(clone.metaWindow, time);
+        Main.overview.activateWindow(clone.metaWindow, time);
     },
 
     _removeSelf : function(actor, event) {
@@ -834,7 +839,7 @@ Workspace.prototype = {
 
             // Set a hint on the Mutter.Window so its initial position
             // in the new workspace will be correct
-            win._overlayHint = {
+            win._overviewHint = {
                 x: actor.x,
                 y: actor.y,
                 scale: actor.scale_x
@@ -845,9 +850,9 @@ Workspace.prototype = {
                                                  false, // don't create workspace
                                                  time);
             return true;
-        } else if (source instanceof GenericDisplay.GenericDisplayItem || source instanceof AppDisplay.WellDisplayItem) {
+        } else if (source.shellWorkspaceLaunch) {
             this._metaWorkspace.activate(time);
-            source.launch();
+            source.shellWorkspaceLaunch();
             return true;
         }
 
@@ -888,27 +893,30 @@ Workspaces.prototype = {
             }
         }
         activeWorkspace.actor.raise_top();
-        this._positionWorkspaces(global, activeWorkspace);
+        this._positionWorkspaces(global);
+
+        // Save the button size as a global variable so we can us it to create 
+        // matching button sizes for workspace remove buttons.
+        buttonSize = addButtonSize;
 
         // Create (+) button
-        buttonSize = addButtonSize;
-        this.addButton = new Clutter.Texture({ x: addButtonX,
-                                                y: addButtonY,
-                                                width: buttonSize,
-                                                height: buttonSize,
-                                                reactive: true
-                                              });
-        this.addButton.set_from_file(global.imagedir + "add-workspace.svg");
-        this.addButton.connect('button-release-event', this._appendNewWorkspace);
-        this.actor.add_actor(this.addButton);
-        this.addButton.lower_bottom();
+        this.addButton = new AddWorkspaceButton(addButtonSize, addButtonX, addButtonY, Lang.bind(this, this._addWorkspaceAcceptDrop));
+        this.addButton.actor.connect('button-release-event', Lang.bind(this, this._appendNewWorkspace));
+        this.actor.add_actor(this.addButton.actor);
+        this.addButton.actor.lower_bottom();
 
         let lastWorkspace = this._workspaces[this._workspaces.length - 1];
         lastWorkspace.updateRemovable(true);
 
-        // Position/scale the desktop windows and their children
-        for (let w = 0; w < this._workspaces.length; w++)
-            this._workspaces[w].zoomToOverlay();
+        // Position/scale the desktop windows and their children after the
+        // workspaces have been created. This cannot be done first because
+        // window movement depends on the Workspaces object being accessible
+        // as an Overview member.
+        Main.overview.connect('showing',
+                             Lang.bind(this, function() {
+            for (let w = 0; w < this._workspaces.length; w++)
+                this._workspaces[w].zoomToOverview();
+        }));
 
         // Track changes to the number of workspaces
         this._nWorkspacesNotifyId =
@@ -928,8 +936,8 @@ Workspaces.prototype = {
         return null;
     },
 
-    // Should only be called from active overlay context
-    activateWindowFromOverlay: function (metaWindow, time) {
+    // Should only be called from active Overview context
+    activateWindowFromOverview: function (metaWindow, time) {
         let global = Shell.Global.get();
         let activeWorkspaceNum = global.screen.get_active_workspace_index();
         let windowWorkspaceNum = metaWindow.get_workspace().index();
@@ -945,27 +953,16 @@ Workspaces.prototype = {
         }
     },
 
-    // Updates position of the workspaces display based on the new coordinates.
-    // Preserves the old value for the coordinate, if the passed value is null.
-    updatePosition : function(x, y) {
-        if (x != null)
-            this._x = x;
-        if (y != null)
-            this._y = y;
-
-        this._updateInOverlay();
-    },
-
     hide : function() {
         let global = Shell.Global.get();
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
         let activeWorkspace = this._workspaces[activeWorkspaceIndex];
 
-        this._positionWorkspaces(global, activeWorkspace);
+        this._positionWorkspaces(global);
         activeWorkspace.actor.raise_top();
 
         for (let w = 0; w < this._workspaces.length; w++)
-            this._workspaces[w].zoomFromOverlay();
+            this._workspaces[w].zoomFromOverview();
     },
 
     destroy : function() {
@@ -982,35 +979,17 @@ Workspaces.prototype = {
         global.window_manager.disconnect(this._switchWorkspaceNotifyId);
     },
 
-    getFullSizeX : function() {
-        return this._workspaces[0].fullSizeX;
+    getScale : function() {
+        return this._workspaces[0].scale;
     },
 
-    // If j-th workspace in the i-th row is active, returns the full width
-    // of j workspaces including empty space if i = 1, or the width of one
-    // workspace.
-    // Used in overlay.js to determine when it is ok to remove the sideshow
-    // during animations for entering and leaving the overlay.
-    getWidthToTopActiveWorkspace : function() {
+    // Get the grid position of the active workspace.
+    getActiveWorkspacePosition : function() {
         let global = Shell.Global.get();
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
         let activeWorkspace = this._workspaces[activeWorkspaceIndex];
 
-        if (activeWorkspace.gridRow == 0)
-            return (activeWorkspace.gridCol + 1) * global.screen_width + activeWorkspace.gridCol * GRID_SPACING;
-        else
-            return global.screen_width;
-    },
-
-    // Updates the workspaces display based on the current dimensions and position.
-    _updateInOverlay : function() {
-        let global = Shell.Global.get();  
-  
-        this._positionWorkspaces(global);
-
-        // Position/scale the desktop windows and their children
-        for (let w = 0; w < this._workspaces.length; w++)
-            this._workspaces[w].updateInOverlay();
+        return [activeWorkspace.gridX, activeWorkspace.gridY];
     },
 
     // Assign grid positions to workspaces. We can't just do a simple
@@ -1023,12 +1002,7 @@ Workspaces.prototype = {
     // first row.)
     //
     // FIXME: need to make the metacity internal layout agree with this!
-    _positionWorkspaces : function(global, activeWorkspace) {
-        if (!activeWorkspace) {
-            let activeWorkspaceIndex = global.screen.get_active_workspace_index();
-            activeWorkspace = this._workspaces[activeWorkspaceIndex];
-        }
-
+    _positionWorkspaces : function(global) {
         let gridWidth = Math.ceil(Math.sqrt(this._workspaces.length));
         let gridHeight = Math.ceil(this._workspaces.length / gridWidth);
 
@@ -1063,14 +1037,6 @@ Workspaces.prototype = {
                 }
             }
         }
-
-        // Now figure out their full-size coordinates
-        for (let w = 0; w < this._workspaces.length; w++) {
-            let workspace = this._workspaces[w];
-
-            workspace.fullSizeX = (workspace.gridCol - activeWorkspace.gridCol) * (global.screen_width + GRID_SPACING);
-            workspace.fullSizeY = (workspace.gridRow - activeWorkspace.gridRow) * (global.screen_height + GRID_SPACING);
-        }
     },
 
     _workspacesChanged : function() {
@@ -1141,6 +1107,11 @@ Workspaces.prototype = {
 
             // FIXME: deal with windows on the lost workspaces
         }
+
+        // Reset the selection state; if we went from > 1 workspace to 1,
+        // this has the side effect of removing the frame border
+        let activeIndex = global.screen.get_active_workspace_index();
+        this._workspaces[activeIndex].setSelected(true);
     },
 
     _activeWorkspaceChanged : function(wm, from, to, direction) {
@@ -1154,10 +1125,39 @@ Workspaces.prototype = {
         this.actor.add_actor(workspace.actor);
     },
 
-    _appendNewWorkspace : function(actor, event) {
+    _appendNewWorkspace : function() {
         let global = Shell.Global.get();
 
-        global.screen.append_new_workspace(false, event.get_time());
+        global.screen.append_new_workspace(false, global.screen.get_display().get_current_time());
+    },
+
+    // Creates a new workspace and drops the dropActor there
+    _addWorkspaceAcceptDrop : function(source, dropActor, x, y, time) {
+        this._appendNewWorkspace();
+        return this._workspaces[this._workspaces.length - 1].acceptDrop(source, dropActor, x, y, time);
+    }
+};
+
+function AddWorkspaceButton(buttonSize, buttonX, buttonY, acceptDropCallback) {
+    this._init(buttonSize, buttonX, buttonY, acceptDropCallback);
+}
+
+AddWorkspaceButton.prototype = {
+    _init : function(buttonSize, buttonX, buttonY, acceptDropCallback) {
+        this.actor = new Clutter.Texture({ x: buttonX,
+                                           y: buttonY,
+                                           width: buttonSize,
+                                           height: buttonSize,
+                                           reactive: true });
+        this._acceptDropCallback = acceptDropCallback;
+        let global = Shell.Global.get();
+        this.actor._delegate = this;
+        this.actor.set_from_file(global.imagedir + 'add-workspace.svg');
+    },
+
+    // Draggable target interface
+    acceptDrop : function(source, actor, x, y, time) {
+        return this._acceptDropCallback(source, actor, x, y, time);
     }
 };
 
@@ -1167,28 +1167,49 @@ Workspaces.prototype = {
 Tweener.registerSpecialPropertyModifier("workspace_relative", _workspaceRelativeModifier, _workspaceRelativeGet);
 
 function _workspaceRelativeModifier(workspace) {
-    let endX, endY;
+    let [startX, startY] = Main.overview.getPosition();
+    let overviewPosX, overviewPosY, overviewScale;
 
     if (!workspace)
         return [];
 
-    if (workspace.leavingOverlay) {
-        endX = workspace.fullSizeX;
-        endY = workspace.fullSizeY;        
+    if (workspace.leavingOverview) {
+        let [zoomedInX, zoomedInY] = Main.overview.getZoomedInPosition();
+        overviewPosX = { begin: startX, end: zoomedInX };
+        overviewPosY = { begin: startY, end: zoomedInY };
+        overviewScale = { begin: Main.overview.getScale(),
+                          end: Main.overview.getZoomedInScale() };
     } else {
-        endX = workspace.gridX;
-        endY = workspace.gridY;
+        overviewPosX = { begin: startX, end: 0 };
+        overviewPosY = { begin: startY, end: 0 };
+        overviewScale = { begin: Main.overview.getScale(), end: 1 };
     }
 
     return [ { name: "x",
-               parameters: { begin: workspace.actor.x, end: endX,
-                             cur: function() { return workspace.actor.x; } } },
+               parameters: { workspacePos: workspace.gridX,
+                             overviewPos: overviewPosX,
+                             overviewScale: overviewScale } },
              { name: "y",
-               parameters: { begin: workspace.actor.y, end: endY,
-                             cur: function() { return workspace.actor.y; } } }
+               parameters: { workspacePos: workspace.gridY,
+                             overviewPos: overviewPosY,
+                             overviewScale: overviewScale } }
            ];
 }
 
 function _workspaceRelativeGet(begin, end, time, params) {
-    return (begin + params.begin) + time * (end + params.end - (begin + params.begin)) - params.cur();
+    let curOverviewPos = (1 - time) * params.overviewPos.begin +
+                         time * params.overviewPos.end;
+    let curOverviewScale = (1 - time) * params.overviewScale.begin +
+                           time * params.overviewScale.end;
+
+    // Calculate the screen position of the window.
+    let screen = (1 - time) *
+                 ((begin + params.workspacePos) * params.overviewScale.begin +
+                  params.overviewPos.begin) +
+                 time *
+                 ((end + params.workspacePos) * params.overviewScale.end +
+                 params.overviewPos.end);
+
+    // Return the workspace coordinates.
+    return (screen - curOverviewPos) / curOverviewScale - params.workspacePos;
 }
